import AutoRPE.UtilsRPE.BasicFunctions as BasicFunctions
import AutoRPE.UtilsRPE.BasicStructures as BasicStructures
import AutoRPE.UtilsRPE.AttributeParser as AttributeParser
import AutoRPE.UtilsRPE.VariablePrecision as VariablePrecision
import AutoRPE.UtilsRPE.RegexPattern as RegexPattern
import AutoRPE.UtilsRPE.Error as Error
import re


def update_precision_declaration(variable, modules, target_precision="dp"):
    """
    Find the file where a variable is declared and replace the declaration precision specification with target_precision
    """
    var_module = variable.procedure.module.name
    # This is the module where the variables appear declared once the code is preprocessed
    preprocessed_module = [m for m in modules if m.module_name == var_module]

    if len(preprocessed_module) > 1 or not len(preprocessed_module):
        raise Error.ExceptionNotManaged

    preprocessed_module = preprocessed_module[0]
    included_files = find_included_files(preprocessed_module, modules)

    # Find the variable original module and line declaration, merge the line if is split
    if variable.is_retval:
        print("The variable %s is actually a function. Skip" % variable.name)
        return

    original_module, index_to_modify = find_and_merge_variable_declaration_line(variable, preprocessed_module, included_files)
    line_to_modify = original_module.lines[index_to_modify]

    # Check that precision declaration has a different precision wrt target_precision
    original_precision = re.search(RegexPattern.precision_of_real_declaration, line_to_modify, re.I)

    if not original_precision:
        raise Error.ExceptionNotManaged("Precision was not found for this declaration")
    original_precision = original_precision.group(3)
    # Check the precision is correct
    try:
        VariablePrecision.lookup_table[original_precision]
    except KeyError:
        raise Error.ExceptionNotManaged("Something went wrong, %s is not a real precision" % original_precision)

    # Avoid replacing variables whose precision was already of the correct type
    # TODO: these variables should not enter the analysis from the beginning
    if original_precision == target_precision:
        return

    original_module.lines[index_to_modify] = replace_real_precision(variable, line_to_modify,
                                                                    target_precision=target_precision)
    original_module.rebuild_text()


def replace_real_precision(variable, declaration_line, target_precision="dp"):
    """This function takes a declaration line (string) and updates it with the new type
    If the declaration contains more variables it keeps the other variables with the original attributes."""
    import warnings
    declaration_line_without_comments = BasicFunctions.remove_comments(declaration_line).strip()

    attributes, arguments = declaration_line_without_comments.split("::")
    variable_name = AttributeParser.get_variable_name(arguments)

    # Real statement pattern
    pattern = "real *\(([^\)]*)\)"

    if len(variable_name) == 1:
        m = re.search(pattern, declaration_line_without_comments, re.I)
        if m:
            real_statement = m.group(0)
            precision_specification = m.group(1)
            new_real_statement = real_statement.replace(precision_specification, target_precision)
            new_declaration_line_without_comments = \
                declaration_line_without_comments.replace(real_statement, new_real_statement)
            new_declaration_line = \
                declaration_line.replace(declaration_line_without_comments, new_declaration_line_without_comments)
        else:
            warnings.warn("Problems finding real statement:\n%s" % declaration_line)
            new_declaration_line = "! WARNING!! the declaration hasn't been properly fixed!\n" + \
                                   declaration_line
        return new_declaration_line
    else:
        m = re.search(pattern, attributes, re.I)
        if m:
            real_statement = m.group(0)
            precision_specification = m.group(1)
            new_real_statement = real_statement.replace(precision_specification, target_precision)
            new_attributes = attributes.replace(real_statement, new_real_statement)
            additional_declaration = "%s :: %s" % (new_attributes, variable.name)
            m = re.search(r"(.*\b)%s(\b.*)" % variable.name, declaration_line_without_comments)
            if m:

                declaration_line_without_selected_var = "".join(m.groups(0))
            else:
                warnings.warn("WTF?")

            # Comma patterns to remove the extra comma
            beginning_pat = "::( *\,)"
            beginning_pat_substitution = "::"
            middle_pat = ",( *,)"
            middle_pat_sub = ","
            last_pat = "(, *)$"
            last_pat_sub = ""

            for pat, sub in [(beginning_pat, beginning_pat_substitution), (middle_pat, middle_pat_sub),
                             (last_pat, last_pat_sub)]:
                m = re.search(pat, declaration_line_without_selected_var)
                if m:
                    new_declaration_line_without_comments = re.sub(pat, sub, declaration_line_without_selected_var)
                    new_declaration_line = declaration_line.replace(declaration_line_without_comments,
                                                                    new_declaration_line_without_comments)
                    break
        else:
            pass

        # Adding proper indentation
        leading_spaces = (len(new_declaration_line) - len(new_declaration_line.lstrip()))
        new_declaration_line = "%s%s\n%s" % (
            leading_spaces * " ", additional_declaration.lstrip(), new_declaration_line)

        return new_declaration_line


def find_and_merge_variable_declaration_line(variable, module, additional_modules):
    """
    Given a variable and a module:
              1) Search for the variable.subprogram declaration
              2) Among the declarations search the variable.name
              3) If the variable declaration is split, merge the line and update both the module and the index
    Returns the index
    """

    # The var is a global variable, declared at the beginning of a module
    if variable.procedure.name == "main":
        start, end = find_main_declarations_index(module)
    # The variable is member of a structure
    elif isinstance(variable.procedure, BasicStructures.DerivedType):
        start, end = find_structure_declarations_index(variable.procedure, module)
    # The variable is a private var of a subprogram
    else:
        try:
            start, end = find_subprogram_declarations_index(variable.procedure, module)
        except Error.ProcedureNotFound:
            module, start, end = search_subprogram_in_additional_modules(variable.procedure, additional_modules)

    declaration_index = None
    for index, line in enumerate(module.lines[start:end + 1]):
        # Skip lines that contains just comments
        line = BasicFunctions.remove_comments(line)
        if not line.strip():
            continue
        # The firs match will be declaration, since we are checking only declaration lines
        if re.search(RegexPattern.name_occurrence % variable.name, line):
            # Check declaration is on the same line of precision specification
            declaration_index = start + index
            if not line.count('::') or line.count("&"):
                # It rebuilds split line inside the module object and return merged line, and updated index
                declaration_index = rebuild_split_line(declaration_index, module)
            break

    if declaration_index is None:
        # The var can be declared in some .h90 included file
        if not additional_modules:
            raise Error.ExceptionNotManaged("Check for variables in include files, probably recursively")
        for additional_mod in additional_modules:
            module, declaration_index
        raise Error.ExceptionNotManaged("Line not found")

    # Return original module object  and index of the declaration
    return module, declaration_index


def search_subprogram_in_additional_modules(procedure, additional_modules):
    if not len(additional_modules):
        raise Error.ProcedureNotFound
    for additional_m in additional_modules:
        try:
            start, end = find_subprogram_declarations_index(procedure, additional_m)
        except Error.ProcedureNotFound:
            continue
        return additional_m, start, end
    raise Error.ProcedureNotFound


def find_main_declarations_index(module):
    """Find start and end indexes of variables declaration in main """
    main_start = -1
    # In case in main there is no subprogram declaration, just variables
    main_end = len(module.lines) - 1
    for index, line in enumerate(module.lines):
        line = BasicFunctions.remove_comments(line)
        # Start of the declaration section
        if line.count("::") and main_start < 0:
            main_start = index
        # Subprogram declaration start
        if line.lower().count('contains'):
            main_end = index - 1
            break
    # If no variable declaration was found
    if main_start < 0:
        return [-1, -1]
    return [main_start, main_end]


def find_structure_declarations_index(structure, module):
    """Find start and end indexes of variables declaration of custom data type """
    start = -1

    begin_pattern = RegexPattern.type_name_begin % structure.name
    end_pattern = RegexPattern.end_type_statement
    for index, line in enumerate(module.lines):
        line = BasicFunctions.remove_comments(line)
        # Start of the declaration section, just if is has not been found yet (avoid match with the end pattern)
        if start < 0 and re.search(begin_pattern, line, re.I):
            start = index
        # Subprogram declaration start
        if start > 0 and re.search(end_pattern, line, re.I):
            end = index - 1
            break
    # Structure was not found
    if start < 0:
        return [-1, -1]
    return [start, end]


def find_subprogram_declarations_index(subprogram, module=None):
    """Find start and end indexes of variables declaration in subprogram code """
    indices = []
    if module is None:
        module = subprogram.module
    begin_pattern = RegexPattern.subprogram_name_declaration % subprogram.name
    end_pattern = RegexPattern.subprogram_name_end_declaration % subprogram.name
    # First find the subprogram body inside the module
    for index, line in enumerate(module.lines):
        # If it's the start or end of the subprogram
        line = BasicFunctions.remove_comments(line)
        if re.search(begin_pattern, line, re.I):
            indices.append(index)
            # End of the program was hit
            if re.search(end_pattern, line, re.I):
                break
    if not len(indices):
        raise Error.ProcedureNotFound

    start = -1
    end = 0
    # The search for declaration
    for index, line in enumerate(module.lines[indices[0]:indices[1] + 1]):

        line = BasicFunctions.remove_comments(line)
        # Skip only comment lines
        if not line.strip():
            continue
        # Skip until first declaration is found
        if start < 0:
            if not line.count('::'):
                continue
        # First declaration found
        if start < 0:
            start = indices[0] + index

        # There can be anything but comments among var declarations (but they can be split)
        # if is not a declaration means the body of the function begun
        if not (line.count('::') or previous_line_is_split(indices[0] + index, module.lines)) and start > 0:
            end = indices[0] + index - 1
            break

    # Return variables declaration line indexes
    return start, end


def previous_line_is_split(index, module_lines):
    """Check if last previous line that was not a comment was a split line"""
    index -= 1
    while not BasicFunctions.remove_comments(module_lines[index]).strip():
        index -= 1
    if module_lines[index].count("&"):
        return True
    return False


def rebuild_split_line(line_index, original_module):
    """Takes a line that can be either split upward or downward, rebuilt it and modify the original module text"""
    # The variable is declared on the first line, but the line is split downward
    if original_module.lines[line_index].count("::"):
        merge_line_in_file(line_index, original_module)
    # The variable is one or more lines down the declaration, go up until the declaration is found
    while not original_module.lines[line_index].count("::"):
        line_index -= 1
    # Now merge lines downwards
    merge_line_in_file(line_index, original_module)
    return line_index


def merge_line_in_file(line_index, module):
    """
    Given a line index corresponding to a split line, copies all the  pieces up to the line index and remove empty lines
    """
    line = BasicFunctions.merge_line_range(module.lines, line_index)
    module.lines[line_index] = line
    # Remove lines that will be empty now
    for j in range(1, BasicFunctions.merge_line_range.number):
        module.lines[line_index + j] = ""
    # Rebuilt text and lines
    module.rebuild_text()


def find_included_files(original_module, modules):
    # The var can be declared in some .h90 included file
    additional_modules = []
    # Search for the include pattern and return the h90 file name
    additional_module_names = [re.search(r"^#\s*include\s+\"(\w+.h90)\"", l, re.I).group(1)
                               for l in original_module.lines if re.search(r"^#\s*include", l, re.I)]
    # If some h90 file is included, create a list with corresponding modules object
    if additional_module_names:
        additional_modules = [m for m in modules if m.filename in additional_module_names]
    return additional_modules
